#include "Main.h"

void MainApp::processMovement(const vec3 &newPosition){
	vec3 pos = newPosition;

	coll.pushSphere(pos, 32);

	position = pos;
}

void MainApp::resetCamera(){
	position = vec3(-460, -200, -60);
	wx =  0;
	wy = -0.76f;
	wz =  0;
}

void MainApp::selectPixelFormat(PixelFormat &pf){
	pf.accumBits = 64;
}

void MainApp::initMenu(){
	Menu *menu = menuSystem->getMainMenu();

	static char *options[] = { "1", "2", "4", "11" };

	sampleMode = 3;
	menu->addMenuItem("Samples: ", &sampleMode, INPUT_INTEGER, 0, 3, options);

	App::initMenu();
}

bool MainApp::init(){
	unsigned int i, j;

	model = new OpenGLModel();
	model->loadFromFile("../Models/Corridor/Map.hmdl");

	for (i = 0; i < model->getBatchCount(); i++){
		model->getBatch(i)->insertAttribute(ATT_TEXCOORD, ATT_FLOAT, 2, 1);
	}

	TexturePacker texPacker;
	for (i = 0; i < model->getBatchCount(); i++){
		Batch *batch = model->getBatch(i);
		Vertex *src = (Vertex *) batch->getVertices();
		unsigned short *inds = (unsigned short *) batch->getIndices();

		for (j = 0; j < batch->getIndexCount(); j += 6){
			float w = length(src[inds[j + 1]].pos - src[inds[j]].pos) / 18;
			float h = length(src[inds[j + 5]].pos - src[inds[j]].pos) / 18;

			if (w < 8) w = 8;
			if (h < 8) h = 8;

			unsigned int tw = (int) w;
			unsigned int th = (int) h;

			tw = (tw + 4) & ~7;
			th = (th + 4) & ~7;

			texPacker.addRectangle(tw, th);
		}
	}

	unsigned int tw = 256;
	unsigned int th = 256;

	if (!texPacker.assignCoords(&tw, &th)){
#ifdef _WIN32
		MessageBox(NULL, "Lightmap too small", "Error", MB_OK);
#endif
		return false;
	}

	tw = getUpperPowerOfTwo(tw);
	th = getUpperPowerOfTwo(th);

	unsigned int index = 0;
	for (i = 0; i < model->getBatchCount(); i++){
		Batch *batch = model->getBatch(i);
		Vertex *src = (Vertex *) batch->getVertices();
		unsigned short *inds = (unsigned short *) batch->getIndices();

		for (j = 0; j < batch->getIndexCount(); j += 6){
			TextureRectangle *rect = texPacker.getRectangle(index);

			float x0 = float(rect->x + 0.5f) / tw;
			float x1 = float(rect->x + rect->width - 0.5f) / tw;
			float y0 = float(rect->y + 0.5f) / th;
			float y1 = float(rect->y + rect->height - 0.5f) / th;

			src[inds[j    ]].texCoord2 = vec2(x0, y0);
			src[inds[j + 1]].texCoord2 = vec2(x1, y0);
			src[inds[j + 2]].texCoord2 = vec2(x1, y1);
			src[inds[j + 5]].texCoord2 = vec2(x0, y1);

			coll.addTriangle(src[inds[j]].pos, src[inds[j + 1]].pos, src[inds[j + 2]].pos);
			coll.addTriangle(src[inds[j + 3]].pos, src[inds[j + 4]].pos, src[inds[j + 5]].pos);

			index++;
		}
	}
	coll.build();

	wa = true;
	return true;
}

bool MainApp::exit(){
	delete model;
	return true;
}

bool MainApp::load(){
	if (!GL_ARB_shader_objects_supported || !GL_ARB_vertex_shader_supported || !GL_ARB_fragment_shader_supported || !GL_ARB_shading_language_100_supported){
		addToLog("No GLSL support (GL_ARB_shader_objects, GL_ARB_vertex_shader, GL_ARB_fragment_shader, GL_ARB_shading_language_100)\n");
		return false;
	}

	if ((lighting = renderer->addShader("lighting.shd")) == SHADER_NONE) return false;

	setDefaultFont("../Textures/Fonts/Future.font", "../Textures/Fonts/Future.dds");

	if ((base[0] = renderer->addTexture("../Textures/floor_wood_4.dds")) == TEXTURE_NONE) return false;
	if ((base[1] = renderer->addTexture("../Textures/floor_wood_3.dds")) == TEXTURE_NONE) return false;
	if ((base[2] = renderer->addTexture("../Textures/floor6.dds")) == TEXTURE_NONE) return false;

	if ((lightMap = renderer->addTexture("LightMap.dds", TEX_NOMIPMAPPING)) == TEXTURE_NONE) return false;

	model->uploadToVertexBuffer();

	return true;
}

bool MainApp::unload(){
	return true;
}

void MainApp::drawScene(){
	renderer->changeMask(ALL);
	glClear(GL_DEPTH_BUFFER_BIT);

	for (unsigned int i = 0; i < model->getBatchCount(); i++){
		renderer->setShader(lighting);
		renderer->setTexture("Base", base[i]);
		renderer->setTexture("LightMap", lightMap);
		renderer->apply();

		((OpenGLBatch *) model->getBatch(i))->draw();
	}
}

bool MainApp::drawFrame(){
	glEnable(GL_CULL_FACE);

	glMatrixMode(GL_MODELVIEW);
	modelView = rotateXY(-wx, -wy) * translate(-position);
	glLoadMatrixf(transpose(modelView));

	glMatrixMode(GL_PROJECTION);
	projection = projectionMatrixX(1.5f, float(height) / float(width), 0.1f, 3000);

	/*
		After projection you're in a unit cube, so [-1, 1] in all directions.
		So this is the size of half a pixel.
	*/
	vec2 halfPixel(1.0f / width, 1.0f / height);

	static vec2 samples1[] = {
		vec2(0, 0),
	};

	// Rotated grid sample pattern
	static vec2 samples2[] = {
		vec2(-0.5f,  0.5f),
		vec2( 0.5f, -0.5f),
	};

	// Rotated grid sample pattern
	static vec2 samples4[] = {
		vec2(-0.25f,  0.75f),
		vec2( 0.25f, -0.75f),
		vec2(-0.75f, -0.25f),
		vec2( 0.75f,  0.25f),
	};

	// Sparse grid sample pattern
	static vec2 samples11[] = {
		vec2( 8.0f  / 11.0f, -10.0f / 11.0f),
		vec2(-2.0f  / 11.0f, -8.0f  / 11.0f),
		vec2( 2.0f  / 11.0f, -6.0f  / 11.0f),
		vec2( 10.0f / 11.0f, -4.0f  / 11.0f),
		vec2(-6.0f  / 11.0f, -2.0f  / 11.0f),
		vec2( 0.0f  / 11.0f,  0.0f  / 11.0f),
		vec2( 6.0f  / 11.0f,  2.0f  / 11.0f),
		vec2(-10.0f / 11.0f,  4.0f  / 11.0f),
		vec2( 4.0f  / 11.0f,  6.0f  / 11.0f),
		vec2(-4.0f  / 11.0f,  8.0f  / 11.0f),
		vec2(-8.0f  / 11.0f,  10.0f / 11.0f),
	};


	static int sampleCount[] = { 1, 2, 4, 11 };
	static vec2 *sampleArrays[] = { samples1, samples2, samples4, samples11 };


	int nSamples = sampleCount[sampleMode];
	vec2 *samples = sampleArrays[sampleMode];


	static float lod[] = { 0, -0.5f, -1, -1.73f };
	renderer->changeLod(0, lod[sampleMode]);


	for (int i = 0; i < nSamples; i++){
		// Adjust projection matrix to add the sample offset after projection
		vec3 offset(samples[i] * halfPixel, 0);
		mat4 proj = translate(offset) * projection;

		glLoadMatrixf(transpose(proj));
		drawScene();

		if (sampleMode > 0){
			// Load first rendered image, accumulate the rest
			glAccum(i? GL_ACCUM : GL_LOAD,  1.0f / nSamples);
		}
	}

	if (sampleMode > 0){
		// Return content of accumulation buffer to the frame-buffer
		glAccum(GL_RETURN, 1);
	}

	renderer->changeLod(0, 0);

	return true;
}

App *app = new MainApp();
